import logging
import json
import os
import re
from pydantic.json import pydantic_encoder

from databricks.sdk import (
    WorkspaceClient, AccountClient
)
from typing import List, Dict, Any
from databricks.sdk.errors import NotFound
from databricks.sdk.service.catalog import SecurableType

from policyweaver.models.config import (
    SourceSchema, Source
)
from policyweaver.plugins.databricks.model import (
    DatabricksColumnMask, DatabricksRowFilter, ColumnMaskExtraction, RowFilterDetails, RowFilterDetailGroup, DatabricksUser, DatabricksServicePrincipal, DatabricksGroup,
    DatabricksGroupMember, Account, RowFilterFunctionInfo, TableObject, Workspace, Catalog, Schema, Table,
    Function, FunctionMap, Privilege
)
from policyweaver.core.enum import (
    ColumnMaskType, RowFilterType,
    IamType
)

from policyweaver.core.auth import ServicePrincipal

class DatabricksAPIClient:
    """
    Databricks API Client for fetching account and workspace policies.
    This client uses the Databricks SDK to interact with the Databricks account and workspace
    and retrieve users, service principals, groups, catalogs, schemas, tables, and privileges.
    This class is designed to be used within the Policy Weaver framework to gather and map policies
    from Databricks workspaces and accounts.
    """
    def __init__(self):
        """
        Initializes the Databricks API Client with account and workspace clients.
        Sets up the logger for the client.
        Raises:
            EnvironmentError: If required environment variables are not set.
        """
        self.logger = logging.getLogger("POLICY_WEAVER")

        self.account_client = AccountClient(host="https://accounts.azuredatabricks.net",
                                            client_id=ServicePrincipal.ClientId,
                                            client_secret=os.environ["DBX_ACCOUNT_API_TOKEN"],
                                            account_id=os.environ["DBX_ACCOUNT_ID"])
        
        self.workspace_client = WorkspaceClient(host=os.environ["DBX_HOST"],
                                                azure_tenant_id=ServicePrincipal.TenantId,
                                                azure_client_id=ServicePrincipal.ClientId,
                                                azure_client_secret=ServicePrincipal.ClientSecret)
        
        self.row_filter_func_maps = []

    def __get_account(self) -> Account:
        """
        Fetches the account details including users, service principals, and groups.
        Returns:
            Account: An Account object containing the account ID, users, service principals, and groups.
        """
        account = Account(
            id = self.account_client.api_client.account_id,
            users=self.__get_account_users__(),
            service_principals=self.__get_account_service_principals__(),
            groups=self.__get_account_groups__()
        )

        self.logger.debug(f"DBX Account: {json.dumps(account, default=pydantic_encoder, indent=4)}")

        return account

    def __get_account_users__(self) -> List[DatabricksUser]:
        """
        Retrieves the list of users in the account.
        Returns:
            List[DatabricksUser]: A list of DatabricksUser objects representing the users in the account.
        """
        users = [
            DatabricksUser(
                id=u.id,
                name=u.display_name,
                email="".join([e.value for e in u.emails if e.primary]),
                external_id=u.external_id
            )
            for u in self.account_client.users.list()
        ]

        self.logger.debug(f"DBX ACCOUNT Users: {json.dumps(users, default=pydantic_encoder, indent=4)}")

        return users

    def __get_account_service_principals__(self) -> List[DatabricksServicePrincipal]:
        """
        Retrieves the list of service principals in the account.
        Returns:
            List[DatabricksServicePrincipal]: A list of DatabricksServicePrincipal objects representing
        """
        service_principals = [
            DatabricksServicePrincipal(
                id=s.id,
                name=s.display_name,
                application_id=s.application_id,
                external_id=s.external_id
            )
            for s in self.account_client.service_principals.list()
        ]

        self.logger.debug(f"DBX ACCOUNT Service Principals: {json.dumps(service_principals, default=pydantic_encoder, indent=4)}")

        return service_principals

    @staticmethod
    def get_members(group_id, dbx_groups):
        dbx_group = dbx_groups.get(group_id)
        if not dbx_group:
            print("Group not found")
            return []
        members = list()
        if not ("members" in dbx_group):
            return members

        for member in dbx_group["members"]:
            if "Groups" in member["$ref"]:
                subgroup_id = member["value"]
                subgroup = dbx_groups.get(subgroup_id)
                if subgroup.get("externalId"):
                    member["externalId"] = subgroup["externalId"]
                else:
                    sub_group_members = DatabricksAPIClient.get_members(member["value"], dbx_groups)
                    members.extend(sub_group_members)
        
            member["byGroup"] = dbx_group["displayName"]
            members.append(member)
        return members

    def __get_account_groups__(self) -> List[DatabricksGroup]:
        """
        Retrieves the list of groups in the account.
        Returns:
            List[DatabricksGroup]: A list of DatabricksGroup objects representing the groups in the account.
        """
        groups = []
        dbx_groups = dict()
        for g in self.account_client.groups.list():
            dbx_groups[g.id] = g.as_dict()

        for g in dbx_groups.values():
            if g.get("externalId"):
                group = DatabricksGroup(
                id=g["id"],
                name=g["displayName"],
                members=[],
                external_id=g["externalId"]
            )
            else:
                members = DatabricksAPIClient.get_members(g["id"], dbx_groups)

                group = DatabricksGroup(
                    id=g["id"],
                    name=g["displayName"],
                    members=[]
                )

                member_ids = list()
                for m in members:
                    if m["value"] in member_ids:
                        continue

                    gm = DatabricksGroupMember(
                            id=m["value"],
                            name=m["display"],
                            external_id=m.get("externalId")
                        )

                    if m["$ref"].find("Users") > -1:
                        gm.type = IamType.USER
                    elif m["$ref"].find("ServicePrincipals") > -1:
                        gm.type = IamType.SERVICE_PRINCIPAL
                    else:
                        gm.type = IamType.GROUP

                    group.members.append(gm)
                    member_ids.append(m["value"])

            groups.append(group)

        self.logger.debug(f"DBX ACCOUNT Groups: {json.dumps(groups, default=pydantic_encoder, indent=4)}")
        return groups
    
    def __get_functions__(self) -> List[RowFilterFunctionInfo]:
        """
        Retrieves the list of functions in the workspace.
        Returns:
            List[Function]: A list of Function objects representing the functions in the workspace.
        """
        row_filters = []
        for row_filter_func_map in self.row_filter_func_maps:
            func = self.workspace_client.functions.get(row_filter_func_map.name)
            rf = RowFilterFunctionInfo(
                fullname=func.full_name,
                name=func.name,
                full_data_type=func.full_data_type,
                routine_definition=func.routine_definition
            )
            row_filters.append(rf)

        return row_filters



    def get_workspace_policy_map(self, source: Source) -> tuple[Account, Workspace]:
        """
        Fetches the workspace policy map for a given source.
        Args:
            source (Source): The source object containing the workspace URL, account ID, and API token.
        Returns:
            Tuple[Account, Workspace]: A tuple containing the Account and Workspace objects.
        Raises:
            NotFound: If the catalog specified in the source is not found in the workspace.
        """
        try:
            self.__account = self.__get_account()
            api_catalog = self.workspace_client.catalogs.get(source.name)

            self.logger.debug(f"DBX Policy Export for {api_catalog.name}...")

            self.__workspace = Workspace(
                users=self.__account.users,
                groups=self.__account.groups,
                service_principals=self.__account.service_principals
            )

            self.__workspace.catalog = Catalog(name=api_catalog.name,
                                               column_masks=[], tables_with_masks=[],
                                               row_filters=[], tables_with_rls=[])
            self.__workspace.catalog.schemas = self.__get_catalog_schemas__(api_catalog.name, source.schemas)
            self.__workspace.catalog.privileges = self.__get_privileges__(SecurableType.CATALOG.value, api_catalog.name)

            #self.__workspace.catalog.row_filters = self.__get_functions__()

            self.logger.debug(f"DBX WORKSPACE Policy Map for {api_catalog.name}: {json.dumps(self.__workspace, default=pydantic_encoder, indent=4)}")
            return (self.__account, self.__workspace)
        except NotFound:
            self.logger.error(f"DBX WORKSPACE Catalog {source.name} not found in workspace {source.url}.")
            return None

    def __get_workspace_users__(self) -> List[DatabricksUser]:
        """
        Retrieves the list of users in the workspace.
        Returns:
            List[DatabricksUser]: A list of DatabricksUser objects representing the users in the workspace.
        """
        users = [
            DatabricksUser(
                id=u.id,
                name=u.display_name,
                email="".join([e.value for e in u.emails if e.primary]),
                external_id=u.external_id
            )
            for u in self.workspace_client.users.list()
        ]

        self.logger.debug(f"DBX WORKSPACE Users: {json.dumps(users, default=pydantic_encoder, indent=4)}")

        return users

    def __get_workspace_service_principals__(self) -> List[DatabricksServicePrincipal]:
        """
        Retrieves the list of service principals in the workspace.
        Returns:
            List[DatabricksServicePrincipal]: A list of DatabricksServicePrincipal objects representing
            the service principals in the workspace.
        """
        service_principals = [
            DatabricksServicePrincipal(
                id=s.id,
                name=s.display_name,
                application_id=s.application_id,
                external_id=s.external_id
            )
            for s in self.workspace_client.service_principals.list()
        ]

        self.logger.debug(f"DBX WORKSPACE Service Principals: {json.dumps(service_principals, default=pydantic_encoder, indent=4)}")

        return service_principals

    def __get_workspace_groups__(self) -> List[DatabricksGroup]:
        """
            Retrieves the list of groups in the workspace.
        Returns:
            List[DatabricksGroup]: A list of DatabricksGroup objects representing the groups in the workspace.
        """
        groups = []
        dbx_groups = dict()
        for g in self.workspace_client.groups.list():
            dbx_groups[g.id] = g.as_dict()

        for g in dbx_groups.values():
            if g.get("externalId"):
                group = DatabricksGroup(
                id=g["id"],
                name=g["displayName"],
                members=[],
                external_id=g["externalId"]
            )
            else:
                members = DatabricksAPIClient.get_members(g["id"], dbx_groups)

                group = DatabricksGroup(
                    id=g["id"],
                    name=g["displayName"],
                    members=[]
                )

                member_ids = list()
                for m in members:
                    if m["value"] in member_ids:
                        continue
                    gm = DatabricksGroupMember(
                            id=m["value"],
                            name=m["display"],
                            external_id=m.get("externalId")
                        )

                    if m["$ref"].find("Users") > -1:
                        gm.type = IamType.USER
                    elif m["$ref"].find("ServicePrincipals") > -1:
                        gm.type = IamType.SERVICE_PRINCIPAL
                    else:
                        gm.type = IamType.GROUP

                    group.members.append(gm)
                    member_ids.append(m["value"])

            
            groups.append(group)

        self.logger.debug(f"DBX WORKSPACE Groups: {json.dumps(groups, default=pydantic_encoder, indent=4)}")
        return groups

    def __get_privileges__(self, type:str, name) -> List[Privilege]:
        """
        Retrieves the privileges for a given securable type and name.
        Args:
            type (SecurableType): The type of the securable (e.g., C
            atalog, Schema, Table, Function).
            name (str): The full name of the securable.
        Returns:
            List[Privilege]: A list of Privilege objects representing the privileges assigned to the securable.
        """
        api_privileges = self.workspace_client.grants.get(
            securable_type=type, full_name=name
        )

        privileges =  []

        for p in api_privileges.privilege_assignments:
            privilege = Privilege(principal=p.principal, privileges=[e.value for e in p.privileges])
   
            privileges.append(privilege)

        self.logger.debug(f"DBX WORKSPACE Privileges for {name}-{type}: {json.dumps(privileges, default=pydantic_encoder, indent=4)}")
        return privileges

    def __get_schema_from_list__(self, schema_list, schema) -> Schema:
        if schema_list:
            search = [s for s in schema_list if s.name == schema]

            if search:
                return search[0]

        return None

    def __get_catalog_schemas__(self, catalog: str, schema_filters: List[SourceSchema]) -> List[Schema]:
        """
        Retrieves the schemas for a given catalog, applying any filters specified in the schema_filters.
        Args:
            catalog (str): The name of the catalog to retrieve schemas from.
            schema_filters (List[SourceSchema]): A list of SourceSchema objects containing filters for schemas.
        Returns:
            List[Schema]: A list of Schema objects representing the schemas in the catalog.
        """
        api_schemas = self.workspace_client.schemas.list(catalog_name=catalog)

        if schema_filters:
            self.logger.debug(f"DBX WORKSPACE Policy Export Schema Filters for {catalog}: {json.dumps(schema_filters, default=pydantic_encoder, indent=4)}")
            
            filter = [s.name for s in schema_filters]
            api_schemas = [s for s in api_schemas if s.name in filter]

        schemas = []

        for s in api_schemas:
            if s.name != "information_schema":
                self.logger.debug(f"DBX WORKSPACE Policy Export for schema {catalog}.{s.name}...")
                schema_filter = self.__get_schema_from_list__(schema_filters, s.name)

                tbls = self.__get_schema_tables__(
                    catalog=catalog,
                    schema=s.name,
                    table_filters=None if not schema_filters else schema_filter.tables,
                )

                schemas.append(
                    Schema(
                        name=s.name,
                        tables=tbls,
                        privileges=self.__get_privileges__(
                            SecurableType.SCHEMA.value, s.full_name
                        ),
                        mask_functions=self.__get_column_mask_functions__(
                            catalog, s.name, tbls
                        ),
                    )
                )

        self.logger.debug(f"DBX WORKSPACE Schemas for {catalog}: {json.dumps(schemas, default=pydantic_encoder, indent=4)}")

        return schemas
    
    def __extract_group_from_mask_function__(self, sql_definition: str, column_name: str) -> ColumnMaskExtraction:
        """
        Extracts the group name and mask pattern from a column mask function definition.
        
        Args:
            sql_definition (str): The SQL definition containing is_account_group_member function
        Returns:
            dict: Dictionary containing 'group_name' and 'mask_pattern', or None values if not found
        """
        result = ColumnMaskExtraction(column_mask_type=ColumnMaskType.UNSUPPORTED)
                
        definition = sql_definition.replace("\n", " ").replace("\r", " ").replace(" ", "")
        if not(definition[:8] == "CASEWHEN" and definition[8:31] == "is_account_group_member"):
            self.logger.warning("Unexpected format: does not start with 'CASE WHEN is_account_group_member'")
            return result
        group_name = definition[32:].split(")")[0].replace("'", "").replace('"', '')
        index_ = 32 + len(group_name) + 3
        if definition[index_:index_+4] != "THEN":
            self.logger.warning("Unexpected format: 'THEN' not found where expected.")
            return result
        split_ = definition[index_ + 4 : ].split("ELSE")

        mask = None
        column_name_pos = None
        assigned_value = split_[0]
        if assigned_value[0] in ["'", '"']:
            assigned_value = assigned_value[1:-1]
            mask = assigned_value
        elif column_name == assigned_value:
            column_name_pos = 1

        unassigned_value = split_[1].replace("END", "")
        if unassigned_value[0] in ["'", '"']:
            unassigned_value = unassigned_value[1:-1]
            mask = unassigned_value
        elif column_name == unassigned_value:
            column_name_pos = 2

        result.group_name = group_name
        result.mask_pattern = mask
        if column_name_pos == 1:
            result.column_mask_type = ColumnMaskType.UNMASK_FOR_GROUP
        elif column_name_pos == 2:
            result.column_mask_type = ColumnMaskType.MASK_FOR_GROUP
        else:
            result.column_mask_type = ColumnMaskType.UNSUPPORTED

        
        return result

    def __extract_case_when_logic_row_filter__(self,sql_definition: str) -> RowFilterDetails:
        """
        Parse a CASE WHEN row filter definition to extract group memberships and their values.
        
        Args:
            sql_definition: The SQL CASE statement from a row filter
            
        Returns:
            Dictionary containing:
            - groups: List of dicts with 'group_name' and 'return_value'
            - default_value: The ELSE clause value
            - filter_type: The type of filter logic
        """
        result = RowFilterDetails(groups=[], default_value=None, row_filter_type=RowFilterType.EXPLICIT_GROUP_MEMBERSHIP)
        pattern = r"IS_ACCOUNT_GROUP_MEMBER\('([^']+)'\)THEN([^W]+?)(?:WHEN|ELSE)"
        
        matches = re.findall(pattern, sql_definition, re.IGNORECASE)
        
        for match in matches:
            group_name = match[0]
            # match[1] is string value (like 'T001'), match[2] is boolean value (true/false)
            return_value = match[1] if match[1] else match[2]
            if not group_name or not return_value:
                continue
            
            result.groups.append(RowFilterDetailGroup(
                group_name=group_name,
                return_value=return_value
            ))
        
        # Extract ELSE clause
        else_pattern = r"ELSE(.+?)END"
        else_match = re.search(else_pattern, sql_definition, re.IGNORECASE)
        if else_match:
            result.default_value = else_match.group(1) if else_match.group(1) else else_match.group(2)
        
        if not result.default_value or not result.groups:
            self.logger.warning("Could not fully parse row filter definition.")
            return None
        return result

    def __extract_if_logic_row_filter__(self, definition: str) -> RowFilterDetails:
        """
        Parse an IF row filter definition to extract group membership and their values.
        
        Args:
            sql_definition: The SQL IF statement from a row filter
        Returns:
            Dictionary containing:

        """
        
        group_name = definition[27:].split("'")[1]
        start = f"IF(IS_ACCOUNT_GROUP_MEMBER('{group_name}'),"
        if not definition.startswith(start):
            self.logger.warning("Unexpected format: does not match expected start")
            return result
        condition_for_group, condition_for_others = definition[len(start):].split(",")
        condition_for_group = condition_for_group.strip()
        condition_for_others = condition_for_others.strip()
        if condition_for_others.endswith(")"):
            condition_for_others = condition_for_others[:-1].strip()
        else:
            self.logger.warning("Unexpected format: does not match expected start")
            return result        

        result = RowFilterDetails(
            groups=[RowFilterDetailGroup(
                group_name=group_name,
                return_value=condition_for_group
            )],
            default_value=condition_for_others,
            row_filter_type=RowFilterType.EXPLICIT_GROUP_MEMBERSHIP
        )

        return result


    def __extract_logic_from_row_filter__(self, sql_definition: str) -> RowFilterDetails:
        """
        Extracts the group name and mask pattern from a column mask function definition.
        
        Args:
            sql_definition (str): The SQL definition containing is_account_group_member function

        Returns:
            ColumnMaskExtraction: An object containing the extracted group name and mask pattern
        """

        
        result = RowFilterDetails(row_filter_type=RowFilterType.UNSUPPORTED)
        definition = sql_definition.replace("\n", " ").replace("\r", " ").replace(" ", "") 
        if not definition.startswith("IF(IS_ACCOUNT_GROUP_MEMBER(") and not definition.startswith("CASEWHENIS_ACCOUNT_GROUP_MEMBER("):
            self.logger.warning("Unexpected format: does not start with 'IF(IS_ACCOUNT_GROUP_MEMBER(' or 'CASEWHENIS_ACCOUNT_GROUP_MEMBER('")
            return result

        if definition.startswith("CASEWHENIS_ACCOUNT_GROUP_MEMBER("):
            result_ = self.__extract_case_when_logic_row_filter__(definition)
        else:
            result_ = self.__extract_if_logic_row_filter__(definition)
        
        if result_:
            return result_
        
        return result
    
    def __get_column_mask__(self, catalog_name: str, schema_name: str, table_name: str, column_name: str, func_map: FunctionMap) -> DatabricksColumnMask:
        """Retrieves the column mask for a given function map.
        Args:
            column_name (str): The name of the column to retrieve the mask for.
            func_map (FunctionMap): The FunctionMap object containing the name and columns of the column mask function.
        Returns:
            ColumnMask: A ColumnMask object representing the column mask function.
        """
        func = self.workspace_client.functions.get(func_map.name)
        extraction = self.__extract_group_from_mask_function__(sql_definition=func.routine_definition, column_name=column_name)
        col_mask = DatabricksColumnMask(name=func_map.name,
                              routine_definition=func.routine_definition,
                              column_name=column_name,
                              catalog_name=catalog_name,
                              schema_name=schema_name,
                              table_name=table_name)
        if extraction:
            col_mask.group_name = extraction.group_name
            col_mask.mask_pattern = extraction.mask_pattern
            col_mask.mask_type = extraction.column_mask_type
        return col_mask

    def __get_row_filter__(self, catalog_name: str, schema_name: str, table_name: str, func_map: FunctionMap) -> DatabricksRowFilter:
        """Retrieves the row filter for a given table.
        Args:
            catalog_name (str): The name of the catalog.
            schema_name (str): The name of the schema.
            table_name (str): The name of the table.
            column_name (str): The name of the column to retrieve the row filter for.
            func_map (FunctionMap): The FunctionMap object containing the name and columns of the row filter function.
        Returns:
            DatabricksRowFilter: A DatabricksRowFilter object representing the row filter for the table.
        """
        func = self.workspace_client.functions.get(func_map.name)
        details = self.__extract_logic_from_row_filter__(sql_definition=func.routine_definition)
        row_filter = DatabricksRowFilter(name=func_map.name,
                                         sql=func.routine_definition,
                                         catalog_name=catalog_name,
                                         schema_name=schema_name,
                                         table_name=table_name,
                                         details=details)

        return row_filter

    def __get_schema_tables__(self, catalog: str, schema: str, table_filters: List[str]) -> List[Table]:
        """
        Retrieves the tables for a given catalog and schema, applying any filters specified in the table_filters
        Args:
            catalog (str): The name of the catalog to retrieve tables from.
            schema (str): The name of the schema to retrieve tables from.
            table_filters (List[str]): A list of table names to filter the results.
        Returns:
            List[Table]: A list of Table objects representing the tables in the catalog and schema.
        """
        api_tables = self.workspace_client.tables.list(
            catalog_name=catalog, schema_name=schema
        )

        if table_filters:
            api_tables = [t for t in api_tables if t.name in table_filters]

        tables = []
        for t in api_tables:

            cms = [self.__get_column_mask__(catalog_name=catalog, schema_name=schema, table_name=t.name, column_name=c.name, func_map=FunctionMap(
                                name=c.mask.function_name, columns=c.mask.using_column_names
                            ))
                            for c in t.columns
                            if c.mask
                        ]
            
            rlsfilter = None
            if t.row_filter:
                rlsfilter = self.__get_row_filter__(catalog_name=catalog, schema_name=schema, table_name=t.name, func_map=FunctionMap(
                                    name=t.row_filter.function_name, columns=t.row_filter.input_column_names
                                ))
            if rlsfilter:
                self.__workspace.catalog.row_filters.append(rlsfilter)
                self.__workspace.catalog.tables_with_rls.append(TableObject(catalog_name=catalog,
                                                                            schema_name=schema,
                                                                            table_name=t.name,
                                                                            columns=[c.name for c in t.columns]))
        
            self.__workspace.catalog.column_masks.extend(cms)
            if cms:
                self.__workspace.catalog.tables_with_masks.append(TableObject(catalog_name=catalog,
                                                                              schema_name=schema,
                                                                              table_name=t.name,
                                                                              columns=[c.name for c in t.columns]))

            t_ = Table(
                        name=t.name,
                        row_filter=rlsfilter,
                        column_masks=cms,
                        privileges=self.__get_privileges__(SecurableType.TABLE.value, t.full_name),
                    )
            tables.append(t_)

        self.logger.debug(f"DBX WORKSPACE Tables for {catalog}.{schema}: {json.dumps(tables, default=pydantic_encoder, indent=4)}")

        return tables

    def __get_column_mask_functions__(self, catalog: str, schema: str, tables: List[Table]) -> List[Function]:
        """
        Retrieves the column mask functions for a given catalog and schema.
        Args:
            catalog (str): The name of the catalog to retrieve column mask functions from.
            schema (str): The name of the schema to retrieve column mask functions from.
            tables (List[Table]): A list of Table objects to check for column masks.
        Returns:
            List[Function]: A list of Function objects representing the column mask functions in the catalog and schema.
        """
        inscope = []

        for t in tables:
            if t.row_filter:
                if t.row_filter.name not in inscope:
                    inscope.append(t.row_filter.name)

            if t.column_masks:
                for m in t.column_masks:
                    if m.name not in inscope:
                        inscope.append(m.name)

        functions = [
            Function(
                name=f.full_name,
                sql=f.routine_definition,
                privileges=self.__get_privileges__(SecurableType.FUNCTION.value, f.full_name),
            )
            for f in self.workspace_client.functions.list(
                catalog_name=catalog, schema_name=schema
            )
            if f.full_name in inscope
        ]

        self.logger.debug(f"DBX WORKSPACE Functions for {catalog}.{schema}: {json.dumps(functions, default=pydantic_encoder, indent=4)}") 
        return functions


from logging import config
import logging
import os
from uuid import uuid4
import requests

from policyweaver.models.config import SourceMap
from policyweaver.core.auth import AzureCLIClient
try:
    import notebookutils
except:
    pass

class Configuration:
    """
    Configuration class for managing application settings and environment variables.
    This class provides methods to set and retrieve configuration values,
    including correlation IDs and service principal credentials.
    Example usage:
        config = Configuration()
        config.configure_environment(SourceMap(correlation_id="12345"))
        print(os.environ['CORRELATION_ID'])  # Outputs: 12345
    """
    @staticmethod
    def configure_environment(config:SourceMap):
        """
        Configure the environment with the provided SourceMap configuration.
        This method sets the correlation ID in the environment variables
        and ensures that a unique correlation ID is generated if not provided.
        Args:
            config (SourceMap): The SourceMap instance containing configuration values.
        """
        if not config.correlation_id:
            config.correlation_id = str(uuid4())

        if config.keyvault and config.keyvault.use_key_vault:
            Configuration.retrieve_key_vault_credentials(config)

        os.environ['CORRELATION_ID'] = config.correlation_id

    @staticmethod
    def retrieve_key_vault_credentials(config:SourceMap):
        """
        Retrieve the Key Vault credentials from the configuration.
        This method checks if the Key Vault is enabled and retrieves the necessary credentials.
        Args:
            config (SourceMap): The SourceMap instance containing configuration values.
        Returns:
            dict: A dictionary containing Key Vault credentials if available.
        """
        logger = logging.getLogger("POLICY_WEAVER")

        if not(config.keyvault and config.keyvault.use_key_vault):
            logger.error("Key Vault is not enabled in the configuration.")
            raise ValueError("Key Vault is not enabled in the configuration.")
        if not(config.keyvault.name and config.keyvault.authentication_method):
            logger.error("Key Vault name and authentication method must be provided.")
            raise ValueError("Key Vault name and authentication method must be provided.")
        
        key_vault_name = config.keyvault.name

        if config.keyvault.authentication_method == "azure_cli":
            
            credential = AzureCLIClient()
            credential.initialize()
            
            def get_secret(secret_name, config_name):
                """
                Retrieve a secret from Azure Key Vault using the Azure CLI credentials.
                Args:
                    secret_name (str): The name of the secret to retrieve.
                Returns:
                    str: The value of the secret.
                """
                key_vault_base_url = f"https://{key_vault_name}.vault.azure.net"
                headers = credential.get_token_header(key_vault_base_url)
                
                keyvault_url = f"{key_vault_base_url}/secrets/{secret_name}?api-version=7.4"
                response = requests.get(keyvault_url, headers=headers)
                
                if response.status_code == 200:
                    return response.json()["value"]
                else:
                    logger.warning(f"Failed to retrieve secret '{config_name}' from Key Vault. Using the configured value as fallback.")
                    return secret_name

        elif config.keyvault.authentication_method == "fabric_notebook":
            def get_secret(secret_name, config_name):
                """
                Retrieve a secret from Azure Key Vault using the Fabric Notebook credentials.
                Args:
                    secret_name (str): The name of the secret to retrieve.
                Returns:
                    str: The value of the secret.
                """
                # Assuming notebookutils is available in the environment
                try:
                    return notebookutils.credentials.getSecret(f'https://{key_vault_name}.vault.azure.net/', secret_name)
                except Exception as e:
                    logger.warning(f"Failed to retrieve secret '{config_name}' from Key Vault. Using the configured value as fallback.")
                    return secret_name
            
        else:
            raise ValueError("Unsupported Key Vault authentication method. Use 'azure_cli' or 'fabric_notebook'.")

        config.service_principal.tenant_id = get_secret(config.service_principal.tenant_id, config_name="tenant_id")
        config.service_principal.client_id = get_secret(config.service_principal.client_id, config_name="client_id")
        config.service_principal.client_secret = get_secret(config.service_principal.client_secret, config_name="client_secret")
        if hasattr(config, 'databricks') and config.databricks is not None:
            config.databricks.account_api_token = get_secret(config.databricks.account_api_token, config_name="databricks.account_api_token")
        if hasattr(config, 'snowflake') and config.snowflake is not None:
            config.snowflake.account_name = get_secret(config.snowflake.account_name, config_name="snowflake account_name")
            config.snowflake.user_name = get_secret(config.snowflake.user_name, config_name="snowflake user_name")
            config.snowflake.password = get_secret(config.snowflake.password, config_name="snowflake password")




from typing import Optional, List
from pydantic import Field

import os
import yaml

from policyweaver.models.common import CommonBaseModel
from policyweaver.core.enum import PolicyWeaverConnectorType
from policyweaver.core.exception import PolicyWeaverError

class SourceSchema(CommonBaseModel):
    """
    Represents a schema in a source.
    Attributes:
        name (str): The name of the schema.
        tables (List[str]): A list of table names in the schema.
    """
    name: Optional[str] = Field(alias="name", default=None)
    tables: Optional[List[str]] = Field(alias="tables", default=None)

class Source(CommonBaseModel):
    """
    Represents a source in the Policy Weaver application.
    Attributes:
        id (str): The unique identifier for the source.
        type (PolicyWeaverConnectorType): The type of the source connector.
        name (str): The name of the source.
        schemas (List[SourceSchema]): A list of schemas in the source.
    """
    name: Optional[str] = Field(alias="name", default=None)
    schemas: Optional[List[SourceSchema]] = Field(alias="schemas", default=None)

    def get_schema_list(self) -> List[str]:
        """
        Returns a list of schema names in the source.
        If there are no schemas, returns None.  
        Returns:
            List[str]: A list of schema names if schemas exist, otherwise None.
        """
        if not self.schemas:
            return None

        return [s.name for s in self.schemas]

class ColumnConstraintsConfig(CommonBaseModel):
    """
    Configuration for column constraints in the fabric.
    Attributes:
        columnlevelsecurity (bool): Flag to indicate whether column-level security is enabled.
        fallback (str): The fallback policy, e.g., "deny" or "allow".
    """
    columnlevelsecurity: Optional[bool] = Field(alias="columnlevelsecurity", default=False)
    fallback: Optional[str] = Field(alias="fallback", default="deny")

class RowConstraintsConfig(CommonBaseModel):
    """
    Configuration for row constraints in the fabric.
    Attributes:
        rowlevelsecurity (bool): Flag to indicate whether row-level security is enabled.
        fallback (str): The fallback policy, e.g., "deny" or "allow".
    """
    rowlevelsecurity: Optional[bool] = Field(alias="rowlevelsecurity", default=False)
    fallback: Optional[str] = Field(alias="fallback", default="deny")

class ConstraintsConfig(CommonBaseModel):
    """
    Configuration for fabric constraints in the Policy Weaver application.
    Attributes:
        columns (dict): A dictionary containing column-level security settings.
            - columnlevelsecurity (bool): Flag to indicate whether column-level security is enabled.
            - fallback (str): The fallback policy, e.g., "deny" or "allow".
    """
    columns: Optional[ColumnConstraintsConfig] = Field(alias="columns", default=None)
    rows: Optional[RowConstraintsConfig] = Field(alias="rows", default=None)

class FabricConfig(CommonBaseModel):
    """
    Configuration for the fabric in the Policy Weaver application.
    Attributes:
        tenant_id (str): The Azure tenant ID.
        workspace_id (str): The Azure workspace ID.
        workspace_name (str): The name of the Azure workspace.
        mirror_id (str): The ID of the mirror in the fabric.
        mirror_name (str): The name of the mirror in the fabric.
        fabric_role_suffix (str): The suffix for the fabric role, default is "PWPolicy".
        delete_default_reader_role (bool): Flag to indicate whether to delete the default reader role,
            default is False.
    """
    tenant_id: Optional[str] = Field(alias="tenant_id", default=None)
    workspace_id: Optional[str] = Field(alias="workspace_id", default=None)
    workspace_name: Optional[str] = Field(alias="workspace_name", default=None)
    mirror_id: Optional[str] = Field(alias="mirror_id", default=None)
    mirror_name: Optional[str] = Field(alias="mirror_name", default=None)
    fabric_role_suffix: Optional[str] = Field(alias="fabric_role_suffix", default="PWPolicy")
    delete_default_reader_role: Optional[bool] = Field(alias="delete_default_reader_role", default=False)
    policy_mapping: Optional[str] = Field(alias="policy_mapping", default="table_based")

class ServicePrincipalConfig(CommonBaseModel):
    """
    Configuration for service principal authentication.
    Attributes:
        tenant_id (str): The Azure tenant ID.
        client_id (str): The client ID of the service principal.
        client_secret (str): The client secret of the service principal.
    """
    tenant_id: Optional[str] = Field(alias="tenant_id", default=None)
    client_id: Optional[str] = Field(alias="client_id", default=None)
    client_secret: Optional[str] = Field(alias="client_secret", default=None)

class KeyVaultConfig(CommonBaseModel):
    """
    Configuration for Azure Key Vault.
    Attributes:
        use_key_vault (bool): Flag to indicate whether to use Azure Key Vault.
        name (str): The name of the Key Vault.
        authentication_method (str): The authentication method to use for accessing the Key Vault.
    """
    use_key_vault: Optional[bool] = Field(alias="use_key_vault", default=False)
    name: Optional[str] = Field(alias="name", default=None)
    authentication_method: Optional[str] = Field(alias="authentication_method", default=None)


class CatalogItem(CommonBaseModel):
    """
    Base model for catalog items.
    This model provides common functionality for catalog items such as ID, name, and type.
    Attributes:
        id (str): The unique identifier for the catalog item.
        name (str): The name of the catalog item.
        type (str): The type of the catalog item.
        catalog (str): The catalog to which the item belongs.
        catalog_schema (str): The schema of the catalog.
        table (str): The table associated with the catalog item.
    """
    catalog: Optional[str] = Field(alias="catalog", default=None)
    catalog_schema: Optional[str] = Field(alias="catalog_schema", default=None)
    table: Optional[str] = Field(alias="table", default=None)
    
class SourceMapItem(CatalogItem):
    """
    Represents an item in the source map.
    Attributes:
        id (str): The unique identifier for the source map item.
        name (str): The name of the source map item.
        type (str): The type of the source map item.
        catalog (str): The catalog to which the source map item belongs.
        catalog_schema (str): The schema of the catalog.
        table (str): The table associated with the source map item.
        mirror_table_name (str): The name of the mirror table associated with the source map item.
    """
    mirror_table_name: Optional[str] = Field(alias="mirror_table_name", default=None)

class SourceMap(CommonBaseModel):
    """
    Represents a source map in the Policy Weaver application.
    This model contains configuration for the source map, including application name, correlation ID,
    connector type, source, fabric configuration, service principal configuration, and mapped items.
    Attributes:
        application_name (str): The name of the application using the source map.
        correlation_id (str): A unique identifier for the correlation of operations.
        type (PolicyWeaverConnectorType): The type of the connector used in the source map.
        source (Source): The source from which the policies are mapped.
        fabric (FabricConfig): Configuration for the fabric in which the policies are managed.
        service_principal (ServicePrincipalConfig): Configuration for service principal authentication.
        mapped_items (List[SourceMapItem]): A list of items that are mapped in the source map.
    """
    application_name: Optional[str] = Field(alias="application_name", default="POLICY_WEAVER")
    correlation_id: Optional[str] = Field(alias="correlation_id", default=None)
    type: Optional[PolicyWeaverConnectorType] = Field(alias="type", default=None)
    source: Optional[Source] = Field(alias="source", default=None)
    fabric: Optional[FabricConfig] = Field(alias="fabric", default=None)
    constraints: Optional[ConstraintsConfig] = Field(alias="constraints", default=None)
    service_principal: Optional[ServicePrincipalConfig] = Field(alias="service_principal", default=None)
    mapped_items: Optional[List[SourceMapItem]] = Field(alias="mapped_items", default=None)
    keyvault: Optional[KeyVaultConfig] = Field(alias="keyvault", default=None)

    _default_paths = ['./settings.yaml']

    @classmethod
    def from_yaml(cls, path:str=None) -> 'SourceMap':
        """
        Load a SourceMap instance from a YAML file.
        Args:
            path (str): The path to the YAML file. If None, uses the default paths. 
        Returns:
            SourceMap: An instance of SourceMap loaded from the YAML file.
        Raises:
            PolicyWeaverError: If the YAML file does not exist or cannot be loaded.
        """
        paths = [path] if path else cls._default_paths.default
            
        for p in paths:
            if os.path.exists(p):
                with open(p, 'r', encoding="utf-8") as f:
                    data = yaml.safe_load(f)
                return cls(**data)
        
        raise PolicyWeaverError("Policy Sync settings file not found")

    def __save_to_first_writable_path__(self, path:str=None) -> None:
        """
        Save the SourceMap instance to the first writable path in the provided list or default paths.
        Args:
            path (str): The path to save the YAML file. If None, uses the default paths.
        Raises:
            IOError: If none of the paths are writable.
        """
        paths = [path] if path else self._default_paths

        for p in paths:
            try:
                with open(p, 'w', encoding="utf-8") as f:
                    yaml.safe_dump(self.model_dump(exclude_none=True, exclude_unset=True), f)
                print(f"Settings saved to {p}")
                return
            except IOError:
                print(f"Unable to write to {p}")
        raise IOError(f"None of the paths in {paths} are writable.")

    def to_yaml(self, path:str=None) -> None:
        """
        Save the SourceMap instance to a YAML file.
        Args:
            path (str): The path to save the YAML file. If None, uses the default paths.
        Raises:
            IOError: If none of the paths are writable.
        """
        self.__save_to_first_writable_path__(path)


import requests
import logging

class RestAPIProxy:
    """
    A class to interact with a REST API.
    This class provides methods to perform GET, POST, PUT, and DELETE requests.
    Attributes:
        logger (logging.Logger): Logger instance for logging API interactions.
        base_url (str): The base URL of the REST API.
        headers (dict): Default headers to be used in API requests.
    """
    def __init__(self, base_url, headers=None, weaver_type=None):
        """
        Initializes the RestAPIProxy with a base URL and optional headers.
        Args:
            base_url (str): The base URL of the REST API.
            headers (dict, optional): Default headers to be used in API requests. Defaults to None.
        
        Raises:
            ValueError: If the base URL is not provided.
        """
        self.logger = logging.getLogger("POLICY_WEAVER")
        self.base_url = base_url
        self.headers = headers
        self.weaver_type = weaver_type if weaver_type else "UNKNOWN"

    def get(self, endpoint, params=None, headers=None):
        """
        Performs a GET request to the specified endpoint of the REST API.
        Args:
            endpoint (str): The endpoint to which the GET request is made.
            params (dict, optional): Query parameters to be included in the request. Defaults to None.
            headers (dict, optional): Headers to be included in the request. Defaults to None.
        Returns:
            Response object: The response from the GET request.
        """
        if not headers:
            headers = self.headers

        self.logger.debug(f"REST API PROXY - GET - {self.base_url}/{endpoint} - HEADERS {headers} - PARAMS - {params}")

        response = requests.get(
            f"{self.base_url}/{endpoint}", params=params, headers=headers
        )
        return self._handle_response(response)

    def post(self, endpoint, data=None, json=None, files=None, headers=None):
        """
        Performs a POST request to the specified endpoint of the REST API.  
        Args:
            endpoint (str): The endpoint to which the POST request is made.
            data (dict, optional): Form data to be included in the request. Defaults to None
            json (dict, optional): JSON data to be included in the request. Defaults to None.
            files (dict, optional): Files to be uploaded in the request. Defaults to None.
            headers (dict, optional): Headers to be included in the request. Defaults to None.
        Returns:
            Response object: The response from the POST request.
        """
        if not headers:
            headers = self.headers

        self.logger.debug(f"REST API PROXY - POST - {self.base_url}/{endpoint} - HEADERS {headers} - DATA - {data} - JSON - {json}")

        response = requests.post(
            f"{self.base_url}/{endpoint}",
            data=data,
            json=json,
            files=files,
            headers=headers,
        )
        return self._handle_response(response)

    def put(self, endpoint, data=None, json=None, headers=None):
        """
        Performs a PUT request to the specified endpoint of the REST API.
        Args:
            endpoint (str): The endpoint to which the PUT request is made.
            data (dict, optional): Form data to be included in the request. Defaults to None
            json (dict, optional): JSON data to be included in the request. Defaults to None.
            headers (dict, optional): Headers to be included in the request. Defaults to None.
        Returns:
            Response object: The response from the PUT request."""
        if not headers:
            headers = self.headers

        headers["policyweaver"] = self.weaver_type

        self.logger.debug(f"REST API PROXY - PUT - {self.base_url}/{endpoint} - HEADERS {headers} - DATA - {data} - JSON - {json}") 
        response = requests.put(
            f"{self.base_url}/{endpoint}", data=data, json=json, headers=headers
        )
        return self._handle_response(response)

    def delete(self, endpoint, headers=None):
        """
        Performs a DELETE request to the specified endpoint of the REST API.
        Args:
            endpoint (str): The endpoint to which the DELETE request is made.
            headers (dict, optional): Headers to be included in the request. Defaults to None.  
        Returns:
            Response object: The response from the DELETE request.
        """
        if not headers:
            headers = self.headers

        self.logger.debug(f"REST API PROXY - DELETE - {self.base_url}/{endpoint} - HEADERS {headers}")
        response = requests.delete(f"{self.base_url}/{endpoint}", headers=headers)
        return self._handle_response(response)

    def _handle_response(self, response):
        """
        Handles the response from the REST API. 
        Args:
            response (Response): The response object from the requests library.
        Returns:
            Response object: The response if the status code is 200, 201, or 202.
        Raises:
            HTTPError: If the response status code is not 200, 201, or 202.
        """
        self.logger.debug(f"REST API PROXY - RESPONSE - {response.status_code}")
        if response.status_code in (200, 201, 202):
            return response
        else:
            self.logger.error(f"REST API PROXY - ERROR - {response.status_code} - {response.text}")
            response.raise_for_status()



import json
import logging
import os
from pydantic.json import pydantic_encoder

from typing import List, Tuple

from policyweaver.core.common import PolicyWeaverCore
from policyweaver.plugins.snowflake.model import SnowflakeGrant, SnowflakeMaskingPolicy, SnowflakeRole, SnowflakeSourceMap, SnowflakeUser

from policyweaver.models.export import (
    PermissionScope, PolicyExport, Policy, Permission, PermissionObject, RolePolicy, RolePolicyExport, ColumnConstraint, RowConstraint
)

from policyweaver.core.enum import (
    IamType, PermissionType, PermissionState, PolicyWeaverConnectorType, ColumnMaskType, RowFilterType
)

from policyweaver.core.utility import Utils
from policyweaver.core.common import PolicyWeaverCore
from policyweaver.plugins.snowflake.api import SnowflakeAPIClient

class SnowflakePolicyWeaver(PolicyWeaverCore):
    """
        Snowflake Policy Weaver for Snowflake Databases.
        This class extends the PolicyWeaverCore to implement the mapping of policies
        from Snowflake Database to the Policy Weaver framework.
    """
    sf_read_permissions = ["SELECT", "OWNERSHIP"]
    sf_database_read_prereqs = ["USAGE", "OWNERSHIP"]
    sf_schema_read_prereqs = ["USAGE", "OWNERSHIP"]

    def __init__(self, config:SnowflakeSourceMap) -> None:
        """
        Initializes the SnowflakePolicyWeaver with the provided configuration.
        Args:
            config (SnowflakeSourceMap): The configuration object containing the workspace URL, account ID, and API token.
        Raises:
            ValueError: If the configuration is not of type SnowflakeSourceMap.
        """
        super().__init__(PolicyWeaverConnectorType.SNOWFLAKE, config)

        self.__config_validation(config)
        self.__init_environment(config)
        
        self.workspace = None
        self.account = None
        self.snapshot = {}
        self.api_client = SnowflakeAPIClient()

    def __init_environment(self, config:SnowflakeSourceMap) -> None:
        os.environ["SNOWFLAKE_ACCOUNT"] = config.snowflake.account_name
        os.environ["SNOWFLAKE_USER"] = config.snowflake.user_name
        os.environ["SNOWFLAKE_PASSWORD"] = config.snowflake.password
        os.environ["SNOWFLAKE_WAREHOUSE"] = config.snowflake.warehouse
        if config.snowflake.private_key_file:
            os.environ["SNOWFLAKE_PRIVATE_KEY_FILE"] = config.snowflake.private_key_file

    def __config_validation(self, config:SnowflakeSourceMap) -> None:
        """
        Validates the configuration for the SnowflakePolicyWeaver.
        This method checks if the configuration is of type SnowflakeSourceMap and if all required fields are present.
        Args:
            config (SnowflakeSourceMap): The configuration object to validate.
        Raises:
            ValueError: If the configuration is not of type SnowflakeSourceMap or if any required fields are missing.
        """
        if not config.snowflake:
            raise ValueError("SnowflakeSourceMap configuration is required for SnowflakePolicyWeaver.")

        if not config.snowflake.account_name:
            raise ValueError("Snowflake account name is required in the configuration.")

        if not config.snowflake.user_name:
            raise ValueError("Snowflake user name is required in the configuration.")

        if not config.snowflake.password:
            raise ValueError("Snowflake password is required in the configuration.")

        if not config.snowflake.warehouse:
            raise ValueError("Snowflake warehouse is required in the configuration.")
    
    def __validate_table_grant__(self, grant_to_validate: SnowflakeGrant, role: SnowflakeRole):

        role_grants = [g for g in self.map.grants if g.grantee_name == role.name
                and g.granted_on == "TABLE"
                and g.name == grant_to_validate.name
                and g.table_schema == grant_to_validate.table_schema
                and g.table_catalog == grant_to_validate.table_catalog]
        if any(g.privilege in self.sf_read_permissions for g in role_grants):
            return False
                
        for role_assignment in role.role_assignments:
            role_grants = [g for g in self.map.grants if g.grantee_name == role_assignment.name 
                           and g.granted_on == "TABLE"
                           and g.name == grant_to_validate.name
                           and g.table_schema == grant_to_validate.table_schema
                           and g.table_catalog == grant_to_validate.table_catalog]
            if any(g.privilege in self.sf_read_permissions for g in role_grants):
                return True
        return False


    def __validate_grant__(self, grant_to_validate: SnowflakeGrant):

        grantee_name = grant_to_validate.grantee_name
        database = grant_to_validate.table_catalog
        schema = grant_to_validate.table_schema

        role_user = [role for role in self.map.roles if role.name == grantee_name]
        if not role_user:
            role_user = [user for user in self.map.users if user.name == grantee_name]
        if not role_user:
            return False
        role_user = role_user[0]

        # Check if the grantee has usage permission on database
        db_grants = [grant for grant in self.map.grants if grant.grantee_name == grantee_name 
                     and grant.granted_on == "DATABASE"
                     and grant.name == database]
        has_db_usage = any(grant.privilege in self.sf_database_read_prereqs for grant in db_grants)

        if not has_db_usage:
            # If not check privileges via role assignments
            for role_assignment in role_user.role_assignments:
                db_grants = [grant for grant in self.map.grants if grant.grantee_name == role_assignment.name
                             and grant.granted_on == "DATABASE"
                             and grant.name == database]
                if any(grant.privilege in self.sf_database_read_prereqs for grant in db_grants):
                    has_db_usage = True
                    break

        # Check if the grantee has usage permission on schema
        schema_grants = [grant for grant in self.map.grants if grant.grantee_name == grantee_name
                          and grant.granted_on == "SCHEMA"
                          and grant.name == schema]
        has_schema_usage = any(grant.privilege in self.sf_schema_read_prereqs for grant in schema_grants)
        if not has_schema_usage:
            # If not check privileges via role assignments
            for role_assignment in role_user.role_assignments:
                schema_grants = [grant for grant in self.map.grants if grant.grantee_name == role_assignment.name
                                 and grant.granted_on == "SCHEMA"
                                 and grant.name == schema]
                if any(grant.privilege in self.sf_schema_read_prereqs for grant in schema_grants):
                    has_schema_usage = True
                    break

        return has_db_usage and has_schema_usage

    def __build_special_grants__(self) -> list[SnowflakeGrant]:
        # Build special grants based on column masking policies
        special_grants = []

        for masking_policy in self.map.masking_policies:
            if masking_policy.column_mask_type == ColumnMaskType.UNSUPPORTED:
                continue
            sfg = SnowflakeGrant(table_catalog=masking_policy.database_name,
                                 table_schema=masking_policy.schema_name,
                                 name=masking_policy.table_name)
            for role_name in masking_policy.group_names:
                roles = [r for r in self.map.roles if r.name == role_name]
                if not roles:
                    self.logger.warning(f"Role {role_name} not found in roles list.")
                    continue
                role = roles[0]
                if self.__validate_table_grant__(sfg, role):
                    special_grant = SnowflakeGrant(grantee_name=role_name,
                                                   granted_on="TABLE",
                                                   privilege="SELECT",
                                                   table_catalog=masking_policy.database_name,
                                                   table_schema=masking_policy.schema_name,
                                                   name=masking_policy.table_name)
                    special_grants.append(special_grant)


        for row_filter_policy in self.map.row_access_policies:
            if row_filter_policy.details.row_filter_type == RowFilterType.UNSUPPORTED:
                self.logger.warning(f"Unsupported row filter type for row filter policy {row_filter_policy.name} on {row_filter_policy.database_name}.{row_filter_policy.schema_name}.{row_filter_policy.table_name}")
                self.logger.warning(f"Skipping special grant for this row filter policy.")
                continue

            sfg = SnowflakeGrant(table_catalog=row_filter_policy.database_name,
                                 table_schema=row_filter_policy.schema_name,
                                 name=row_filter_policy.table_name)

            for group_ in row_filter_policy.details.groups:
                roles = [r for r in self.map.roles if r.name == group_.group_name]
                if not roles:
                    self.logger.warning(f"Role {group_.group_name} not found in roles list.")
                    continue
                role = roles[0]
                if self.__validate_table_grant__(sfg, role):
                    special_grant = SnowflakeGrant(grantee_name=group_.group_name,
                                                   granted_on="TABLE",
                                                   privilege="SELECT",
                                                   table_catalog=row_filter_policy.database_name,
                                                   table_schema=row_filter_policy.schema_name,
                                                   name=row_filter_policy.table_name)
                    special_grants.append(special_grant)

        return special_grants

    def __compute_valid_grants__(self) -> list[SnowflakeGrant]:

        valid_grants = []
        grants = self.map.grants
        table_grants = [grant for grant in grants if grant.granted_on == "TABLE" and grant.privilege in self.sf_read_permissions]

        for grant in table_grants:
            if not self.__validate_grant__(grant):
                continue
            valid_grants.append(grant)

        filtered_grants = []
        for vg in valid_grants:
            duplicate = False
            for fg in filtered_grants:
                if (vg.grantee_name == fg.grantee_name and
                    vg.granted_on == fg.granted_on and
                    vg.privilege == fg.privilege and
                    vg.table_catalog == fg.table_catalog and
                    vg.table_schema == fg.table_schema and
                    vg.name == fg.name):
                    duplicate = True
                    break
            if not duplicate:
                filtered_grants.append(vg)

        return filtered_grants

    def __build_permission_object__(self, user: SnowflakeUser) -> PermissionObject:

        if Utils.is_email(user.login_name):
            po = PermissionObject(name=user.id, email=user.login_name, type=IamType.USER)
            return po
        else:
            print(f"Invalid email format for user: {user.id} , {user.login_name}")

    def __get_all_permission_objects__(self, grantee_name: str) -> list[PermissionObject]:

        role_user = [user for user in self.map.users if user.name == grantee_name]
        is_user = True
        if not role_user:
            role_user = [role for role in self.map.roles if role.name == grantee_name]
            is_user = False
        if not role_user:
            raise ValueError(f"Role user not found for grantee: {grantee_name}")
        role_user = role_user[0]

        permission_objects = []
        if is_user:
            permission_object = self.__build_permission_object__(role_user)
            if permission_object:
                permission_objects.append(permission_object)
        else:
            users_added = list()
            for user in role_user.members_user:
                if user.name in users_added:
                    continue
                permission_object = self.__build_permission_object__(user)
                if permission_object:
                    permission_objects.append(permission_object)
                users_added.append(user.name)

        return permission_objects


    def __get_column_constraints__(self, role_assignments: List[SnowflakeRole],
                                   grants: List[SnowflakeGrant]):

        columnconstraints = []
        for grant in grants:
            table_catalog = grant.table_catalog
            table_schema = grant.table_schema
            table_name = grant.name
            matching_mask_policies = [mp for mp in self.map.masking_policies
                                        if mp.database_name == table_catalog and
                                            mp.schema_name == table_schema and
                                            mp.table_name == table_name]

            if not matching_mask_policies:
                continue

        
            table_w_mask = [t for t in self.map.tables_with_masks if t.database_name == table_catalog and
                                                        t.schema_name == table_schema and
                                                        t.table_name == table_name][0]
            all_columns = table_w_mask.column_names
            columns_to_deny = []

            for mp in matching_mask_policies:
                if mp.column_mask_type == ColumnMaskType.UNSUPPORTED:
                    self.logger.warning(f"Unsupported column mask type for masking policy {mp.name} on {mp.database_name}.{mp.schema_name}.{mp.table_name}.{mp.column_name}.")
                    self.logger.warning(f"Using fallback: {self.config.constraints.columns.fallback}")
                    if self.config.constraints.columns.fallback != "grant":
                        columns_to_deny.append(mp.column_name)
                elif mp.column_mask_type == ColumnMaskType.UNMASK_FOR_GROUP:
                    if not any([role for role in role_assignments if role.name in mp.group_names]):
                        columns_to_deny.append(mp.column_name)
                elif mp.column_mask_type == ColumnMaskType.MASK_FOR_GROUP:
                    if any([role for role in role_assignments if role.name in mp.group_names]):
                        columns_to_deny.append(mp.column_name)
                        
            filtered_columns = [col for col in all_columns if col not in columns_to_deny]        
            if len(filtered_columns) == len(all_columns):
                continue
            constraint = ColumnConstraint(catalog_name=mp.database_name,
                                            schema_name=mp.schema_name,
                                            table_name=mp.table_name,
                                            column_names=filtered_columns,
                                            column_effect=PermissionState.GRANT,
                                            column_actions=[PermissionType.SELECT])
            columnconstraints.append(constraint)

        return columnconstraints

    def __get_row_constraints__(self, role_assignments: List[SnowflakeRole], grants: List[SnowflakeGrant]):
        rowconstraints = []
        for grant in grants:
            table_catalog = grant.table_catalog
            table_schema = grant.table_schema
            table_name = grant.name
            matching_raps = [rap for rap in self.map.row_access_policies
                                    if rap.database_name == table_catalog and
                                        rap.schema_name == table_schema and
                                        rap.table_name == table_name]
            
            matching_unsupported_tables = [rap for rap in self.map.unsupported_tables
                                            if rap.database_name == table_catalog and
                                                rap.schema_name == table_schema and
                                                rap.table_name == table_name]
            
            if not matching_raps:
                continue

            table_objects = [(mp.database_name, mp.schema_name, mp.table_name) for mp in matching_raps]
            table_objects = list(set(table_objects))

            for (catalog, schema, table) in table_objects:

                matching_rls_policies_per_table = [mp for mp in matching_raps if mp.database_name == catalog and
                                                    mp.schema_name == schema and
                                                    mp.table_name == table]
                
                matching_unsupported_tables_per_table = [ut for ut in matching_unsupported_tables if ut.database_name == catalog and
                                                            ut.schema_name == schema and
                                                            ut.table_name == table]
                if matching_unsupported_tables_per_table:
                    self.logger.warning(f"Detecting unsupported policies like aggregation, join or projection policy on: {catalog}.{schema}.{table} . Using fallback: {self.config.constraints.rows.fallback}")
                                            
                    if self.config.constraints.rows.fallback != "grant":
                        constraint = RowConstraint(catalog_name=catalog,
                                                schema_name=schema,
                                                table_name=table,
                                                filter_condition="DENYALL")
                        rowconstraints.append(constraint)
                    continue

                for mp in matching_rls_policies_per_table:
                    if mp.details.row_filter_type == RowFilterType.UNSUPPORTED:
                        self.logger.warning(f"Unsupported row filter type for row filter policy {mp.name} on {mp.database_name}.{mp.schema_name}.{mp.table_name}")
                        self.logger.warning(f"Using fallback: {self.config.constraints.rows.fallback}")
                        if self.config.constraints.rows.fallback != "grant":
                            filter_condition = "DENYALL"  # Deny all
                        else:
                            continue
                    elif mp.details.row_filter_type == RowFilterType.EXPLICIT_GROUP_MEMBERSHIP:
                        filter_condition = None
                        for group in mp.details.groups:
                            if group.group_name in [role.name for role in role_assignments]:
                                filter_condition = group.return_value
                                break
                        if not filter_condition:
                            filter_condition = mp.details.default_value
                        if filter_condition == "false":
                            filter_condition = "DENYALL"
                        if filter_condition == "true":
                            continue


                    constraint = RowConstraint(catalog_name=mp.database_name,
                                            schema_name=mp.schema_name,
                                            table_name=mp.table_name,
                                            filter_condition=filter_condition)
                    rowconstraints.append(constraint)
        return rowconstraints


    def _build_role_based_policy__(self, grantee_name: str, grants: list[SnowflakeGrant],
                                   column_security: bool, row_security: bool) -> RolePolicy:

        permission_scopes = []    
        rowconstraints = []
        columnconstraints = []
        for grant in grants:
            table_catalog = grant.table_catalog
            table_schema = grant.table_schema
            table_name = grant.name
            permission_scope = PermissionScope(catalog=table_catalog,
                                               catalog_schema=table_schema,
                                               table=table_name,
                                               name=PermissionType.SELECT,
                                               state=PermissionState.GRANT)
            permission_scopes.append(permission_scope)
    

        roles = [role for role in self.map.roles if role.name == grantee_name]
        if not roles:
            return None
        role = roles[0]
        role_assignments = [role] + role.role_assignments

        if column_security:
            columnconstraints = self.__get_column_constraints__(role_assignments, grants)
        if row_security:
            rowconstraints += self.__get_row_constraints__(role_assignments, grants)

        permission_objects = self.__get_all_permission_objects__(grantee_name)

        if not permission_objects:
            print(f"No valid permission objects found for grantee: {grantee_name}")
            return None
        if not permission_scopes:
            print(f"No valid permission scopes found for grantee: {grantee_name}")
            return None

        policy = RolePolicy(name=grantee_name,
                            permissionscopes=permission_scopes,
                            permissionobjects=permission_objects,
                            columnconstraints=columnconstraints if columnconstraints else None,
                            rowconstraints=rowconstraints if rowconstraints else None)
        return policy

    def __build_role_based_policy_export__(self) -> RolePolicyExport:
        policy_export = RolePolicyExport(source=self.config.source, type=self.config.type,
                                         policies=[])
        
        if not(self.config.constraints and self.config.constraints.columns and self.config.constraints.columns.columnlevelsecurity):
            self.logger.warning("Column level security is not enabled in the config.")
            column_security = False
        else:
            self.logger.info("Column level security is enabled in the config.")
            column_security = True

        if not(self.config.constraints and self.config.constraints.rows and self.config.constraints.rows.rowlevelsecurity):
            self.logger.warning("Row level security is not enabled in the config.")
            row_security = False
        else:
            self.logger.info("Row level security is enabled in the config.")
            row_security = True

        # group grants by granteename
        grants_by_grantee = {}
        for grant in self.valid_grants:
            if grant.grantee_name not in grants_by_grantee:
                grants_by_grantee[grant.grantee_name] = []
            grants_by_grantee[grant.grantee_name].append(grant)

        for grantee_name, grants in grants_by_grantee.items():
            policy = self._build_role_based_policy__(grantee_name, grants, column_security, row_security)
            if policy:
                policy_export.policies.append(policy)

        return policy_export


    @staticmethod
    def __deduplicate_permission_objects__(permission_objects: List[PermissionObject]) -> List[PermissionObject]:
        new_list = []
        for obj in permission_objects:
            if obj.lookup_id not in [o.lookup_id for o in new_list]:
                new_list.append(obj)
        return new_list

    def _build_table_based_policy__(self, table_catalog: str, table_schema: str, table_name: str, grants: List[SnowflakeGrant]) -> Policy:
        policy = Policy(catalog=table_catalog,
                        catalog_schema=table_schema,
                        table=table_name, permissions=[])
        permission_objects = []
        for grant in grants:
            permission_objects.extend(self.__get_all_permission_objects__(grant.grantee_name))

        permission_objects = self.__deduplicate_permission_objects__(permission_objects)

        permission = Permission(name=PermissionType.SELECT, state=PermissionState.GRANT, objects=permission_objects)
        policy.permissions.append(permission)
        return policy

    def __build_table_based_policy_export__(self) -> PolicyExport:
        policy_export = PolicyExport(source=self.config.source, type=self.config.type,
                                     policies=[])

        # group grants by table
        grants_by_table = {}
        for grant in self.valid_grants:
            table_key = (grant.table_catalog, grant.table_schema, grant.name)
            if table_key not in grants_by_table:
                grants_by_table[table_key] = []
            grants_by_table[table_key].append(grant)

        for (table_catalog, table_schema, table_name), grants in grants_by_table.items():
            policy = self._build_table_based_policy__(table_catalog, table_schema, table_name, grants)
            if policy:
                policy_export.policies.append(policy)

        return policy_export

    def map_policy(self, policy_mapping = 'role_based'):
        self.map = self.api_client.__get_database_map__(self.config.source)
        
        # Build special grants based on column masking policies
        self.map.grants.extend(self.__build_special_grants__())

        # Filter out valid grants
        self.valid_grants = self.__compute_valid_grants__()

        if policy_mapping == 'role_based':
            return self.__build_role_based_policy_export__()
        elif policy_mapping == 'table_based':
            return self.__build_table_based_policy_export__()


import logging
import re
import os
from pydantic.json import pydantic_encoder

import snowflake.connector

from typing import List, Tuple

from policyweaver.models.config import (
    SourceSchema, Source
)

from policyweaver.core.enum import ColumnMaskType, RowFilterType

from policyweaver.plugins.snowflake.model import (
    RowFilterDetailGroup,
    RowFilterDetails,
    SnowflakeColumnMask,
    SnowflakeColumnMaskExtraction,
    SnowflakeConnection,
    SnowflakeGrant,
    SnowflakeMaskingPolicy,
    SnowflakeRole,
    SnowflakeRoleMemberMap,
    SnowflakeRowFilter,
    SnowflakeUser,
    SnowflakeDatabaseMap,
    SnowflakeUserOrRole,
    SnowflakeTableWithPolicy
)
from policyweaver.core.enum import (
    IamType
)

from policyweaver.core.auth import ServicePrincipal

class SnowflakeAPIClient:
    """
    Snowflake API Client for fetching account policies.
    This client uses the Snowflake SDK to interact with the Snowflake account
    and retrieve users, databases, schemas, tables, and privileges.
    This class is designed to be used within the Policy Weaver framework to gather and map policies
    from Snowflake workspaces and accounts.
    """
    def __init__(self):
        """
        Initializes the Snowflake API Client with a connection to the Snowflake account.
        Sets up the logger for the client.
        Raises:
            EnvironmentError: If required environment variables are not set.
        """
        self.logger = logging.getLogger("POLICY_WEAVER")

        private_key_file = os.getenv("SNOWFLAKE_PRIVATE_KEY_FILE", None)
        if not private_key_file or not os.path.isfile(private_key_file):
            self.logger.warning("SNOWFLAKE_PRIVATE_KEY_FILE environment variable is not set or the file does not exist. Trying user/password authentication.")
            self.connection = SnowflakeConnection(
                user_name=os.environ["SNOWFLAKE_USER"],
                password=os.environ["SNOWFLAKE_PASSWORD"],
                account_name=os.environ["SNOWFLAKE_ACCOUNT"],
                warehouse=os.environ["SNOWFLAKE_WAREHOUSE"]
            )
        else:
            self.connection = SnowflakeConnection(
                user_name=os.environ["SNOWFLAKE_USER"],
                password=os.environ["SNOWFLAKE_PASSWORD"],
                account_name=os.environ["SNOWFLAKE_ACCOUNT"],
                warehouse=os.environ["SNOWFLAKE_WAREHOUSE"],
                private_key_file=private_key_file
            )

        self.users = None
        self.roles = None
        self.grants = None
        self.role_assignments = None
        self.user_assignments = None
        self.masking_policies = []
        self.tables_with_masks = []



    def __get_snowflake_connection__(self) -> snowflake.connector.SnowflakeConnection:
        """
        Establishes a connection to the Snowflake account.
        Returns:
            snowflake.connector.SnowflakeConnection: A connection object to the Snowflake account.
        Raises:
            EnvironmentError: If required environment variables are not set.
        """

        if self.connection.private_key_file:

            private_key_file = self.connection.private_key_file
            private_key_file_pwd = self.connection.password

            conn_params = {
                'account': self.connection.account_name,
                'user': self.connection.user_name,
                'authenticator': 'SNOWFLAKE_JWT',
                'private_key_file': private_key_file,
                'private_key_file_pwd':private_key_file_pwd,
                'warehouse': self.connection.warehouse,
                'disable_ocsp_checks': True,
                'database': 'SNOWFLAKE',
                'schema': 'ACCOUNT_USAGE'
            }

            ctx = snowflake.connector.connect(**conn_params)
            return ctx
        else:
            return snowflake.connector.connect(
                user=self.connection.user_name,
                password=self.connection.password,
                account=self.connection.account_name,
                warehouse=self.connection.warehouse,
                disable_ocsp_checks=True,
                database="SNOWFLAKE",
                schema="ACCOUNT_USAGE"
            )
        

    def __run_query__(self, query: str, columns: List[str]) -> List[dict]:
        """
        Execute a SQL query against Snowflake and return the results.

        Args:
            query (str): SQL query to execute

        Returns:
            list: Query results as a list of tuples
        """
        with self.__get_snowflake_connection__() as conn:
            with conn.cursor() as cur:
                cur.execute(query)
                results = cur.fetchall()
        return [dict(zip(columns, row)) for row in results]

    def __get_database_map__(self, source: Source) -> SnowflakeDatabaseMap:
        """
        Retrieves the database map from the Snowflake account.
        Returns:
            dict: A dictionary mapping database names to their schemas.
            NotFound: If the catalog specified in the source is not found in the workspace.
        """
        
        # get users

        self.users = self.__get_users__()

        # get roles

        self.roles = self.__get_roles__()

        # get direct and indirect members for each role

        role_memberships = self.__get_role_memberships__()
        for role in self.roles:
            role_name = role.name
            role.members_user = role_memberships[role_name].users
            role.members_role = role_memberships[role_name].roles

        # get user/role to role memberships

        user_role_assignments = self.__get_user_role_assignments__()
        for u in self.users:
            u.role_assignments = user_role_assignments[u.name]
        for r in self.roles:
            r.role_assignments = user_role_assignments[r.name]

        # get grants to roles and users

        self.grants = self.__get_grants__(source)

        # get column masks

        self.__get_column_masks__(source.name)

        # get row access policies
        self.__get_row_access_policies__(source.name)

        self.__get_unsupported_policies__(source.name)

        return SnowflakeDatabaseMap(users=self.users, roles=self.roles,
                                    grants=self.grants,
                                    masking_policies=self.masking_policies,
                                    tables_with_masks=self.tables_with_masks,
                                    row_access_policies=self.row_access_policies,
                                    tables_with_raps=self.tables_with_raps,
                                    unsupported_tables=self.unsupported_tables
                                    )


    def __get_unsupported_policies__(self, database: str) -> None:

        # AGGREGATION POLICIES
        query = f"""                            
                SELECT
                    rap.POLICY_BODY,
                    pr.POLICY_ID,
                    pr.POLICY_NAME,
                    pr.ref_database_name,
                    pr.ref_schema_name,
                    pr.ref_entity_name,
                    pr.ref_column_name
                FROM SNOWFLAKE.ACCOUNT_USAGE.POLICY_REFERENCES pr
                JOIN SNOWFLAKE.ACCOUNT_USAGE.AGGREGATION_POLICIES rap
                    ON pr.POLICY_ID = rap.POLICY_ID
                AND pr.REF_DATABASE_NAME = '{database.upper()}' AND REF_ENTITY_DOMAIN = 'TABLE' AND POLICY_STATUS = 'ACTIVE';

                ;"""

        agg_policies = self.__run_query__(query, columns=["POLICY_BODY", "POLICY_ID",
                                                        "POLICY_NAME", "ref_database_name",
                                                        "ref_schema_name",
                                                        "ref_entity_name",
                                                        "ref_column_name"])
        
        # JOIN POLICIES

        query = f"""
                SELECT
                    rap.POLICY_BODY,
                    pr.POLICY_ID,
                    pr.POLICY_NAME,
                    pr.ref_database_name,
                    pr.ref_schema_name,
                    pr.ref_entity_name,
                    pr.ref_column_name
                FROM SNOWFLAKE.ACCOUNT_USAGE.POLICY_REFERENCES pr
                JOIN SNOWFLAKE.ACCOUNT_USAGE.JOIN_POLICIES rap
                    ON pr.POLICY_ID = rap.POLICY_ID
                AND pr.REF_DATABASE_NAME = '{database.upper()}' AND REF_ENTITY_DOMAIN = 'TABLE' AND POLICY_STATUS = 'ACTIVE';          
                """
        
        join_policies = self.__run_query__(query, columns=["POLICY_BODY", "POLICY_ID",
                                                        "POLICY_NAME", "ref_database_name",
                                                        "ref_schema_name",
                                                        "ref_entity_name",
                                                        "ref_column_name"])

        # PROJECTION POLICIES

        query = f"""                            
                SELECT
                    rap.POLICY_BODY,
                    pr.POLICY_ID,
                    pr.POLICY_NAME,
                    pr.ref_database_name,
                    pr.ref_schema_name,
                    pr.ref_entity_name,
                    pr.ref_column_name
                FROM SNOWFLAKE.ACCOUNT_USAGE.POLICY_REFERENCES pr
                JOIN SNOWFLAKE.ACCOUNT_USAGE.PROJECTION_POLICIES rap
                    ON pr.POLICY_ID = rap.POLICY_ID
                AND pr.REF_DATABASE_NAME = '{database.upper()}' AND REF_ENTITY_DOMAIN = 'TABLE' AND POLICY_STATUS = 'ACTIVE';

                ;"""

        projection_policies = self.__run_query__(query, columns=["POLICY_BODY", "POLICY_ID",
                                                        "POLICY_NAME", "ref_database_name",
                                                        "ref_schema_name",
                                                        "ref_entity_name",
                                                        "ref_column_name"])


        unsupported_policies = agg_policies + join_policies + projection_policies

        unsupported_tables = [SnowflakeTableWithPolicy(database_name=up["ref_database_name"],
                                                       schema_name=up["ref_schema_name"],
                                                       table_name=up["ref_entity_name"],
                                                       column_names=[]) for up in unsupported_policies]

        self.unsupported_tables = unsupported_tables

                                  

    def __get_grants__(self, source: Source) -> List[SnowflakeGrant]:

        query = f"""select   "PRIVILEGE", "GRANTED_ON", "TABLE_CATALOG", "TABLE_SCHEMA", "NAME", "GRANTEE_NAME"
                    from     SNOWFLAKE.ACCOUNT_USAGE.GRANTS_TO_ROLES
                    where    DELETED_ON is null and
                            GRANTED_ON in ('TABLE','SCHEMA','DATABASE') and
                            TABLE_CATALOG = '{source.name}'"""
        
        grants_raw = self.__run_query__(query, columns=["PRIVILEGE", "GRANTED_ON", "TABLE_CATALOG", "TABLE_SCHEMA", "NAME", "GRANTEE_NAME"]) 

        database_grants = [grant for grant in grants_raw if grant["GRANTED_ON"] == "DATABASE"]
        schema_grants = [grant for grant in grants_raw if grant["GRANTED_ON"] == "SCHEMA"]
        table_grants = [grant for grant in grants_raw if grant["GRANTED_ON"] == "TABLE"]

        if source.schemas:
            # filter grants based on provided schemas and tables
            schema_names = [s.name for s in source.schemas]
            filtered_schema_grants = [grant for grant in schema_grants if grant["TABLE_SCHEMA"] in schema_names]
            filtered_table_grants = [grant for grant in table_grants if grant["TABLE_SCHEMA"] in schema_names]
            schema_grants = filtered_schema_grants
            table_grants = filtered_table_grants

            filtered_table_grants = []
            for s in source.schemas:
                if s.tables:
                    filtered_table_grants_per_schema = [grant for grant in table_grants if grant["TABLE_SCHEMA"] == s.name and grant["NAME"] in s.tables]
                else:
                    filtered_table_grants_per_schema = [grant for grant in table_grants if grant["TABLE_SCHEMA"] == s.name]
                filtered_table_grants.extend(filtered_table_grants_per_schema)
            
            table_grants = filtered_table_grants

        grants_raw = database_grants + schema_grants + table_grants

        grants = [SnowflakeGrant(privilege=grant["PRIVILEGE"],
                                 granted_on=grant["GRANTED_ON"],
                                 table_catalog=grant["TABLE_CATALOG"],
                                 table_schema=grant["TABLE_SCHEMA"],
                                 name=grant["NAME"],
                                 grantee_name=grant["GRANTEE_NAME"]) for grant in grants_raw]
        
        return grants
    
    def __get_row_access_policies__(self, database: str) -> None:
        """Retrieves the row access policies for a specific database.
        Args:
            database (str): The name of the database for which to retrieve row access policies.
        """
    
        query = f"""                            
                SELECT
                    rap.POLICY_BODY,
                    pr.POLICY_ID,
                    pr.POLICY_NAME,
                    pr.ref_database_name,
                    pr.ref_schema_name,
                    pr.ref_entity_name,
                    pr.ref_column_name
                FROM SNOWFLAKE.ACCOUNT_USAGE.POLICY_REFERENCES pr
                JOIN SNOWFLAKE.ACCOUNT_USAGE.ROW_ACCESS_POLICIES rap
                    ON pr.POLICY_ID = rap.POLICY_ID
                AND pr.REF_DATABASE_NAME = '{database.upper()}' AND REF_ENTITY_DOMAIN = 'TABLE' AND POLICY_STATUS = 'ACTIVE';

                ;"""

        row_access_policies = self.__run_query__(query, columns=["POLICY_BODY", "POLICY_ID",
                                                        "POLICY_NAME", "ref_database_name",
                                                        "ref_schema_name",
                                                        "ref_entity_name",
                                                        "ref_column_name"])
        
        self.row_access_policies = list()
        self.tables_with_raps = list()
        for rap in row_access_policies:
            extraction = self.__extract_logic_from_row_filter__(rap["POLICY_BODY"])
            mp = SnowflakeRowFilter(id=rap["POLICY_ID"],
                                    name=rap["POLICY_NAME"],
                                    database_name=rap["ref_database_name"],
                                    schema_name=rap["ref_schema_name"],
                                    table_name=rap["ref_entity_name"],
                                    routine_definition=rap["POLICY_BODY"],
                                    details=extraction
                                    )
            self.row_access_policies.append(mp)

            query = f"""SELECT COLUMN_NAME FROM SNOWFLAKE.ACCOUNT_USAGE.COLUMNS WHERE TABLE_CATALOG = '{mp.database_name}' AND TABLE_SCHEMA = '{mp.schema_name}' AND TABLE_NAME = '{mp.table_name}';"""
            columns = self.__run_query__(query, columns=["COLUMN_NAME"])
            column_names = [c["COLUMN_NAME"] for c in columns]
            self.tables_with_raps.append(SnowflakeTableWithPolicy(
                database_name=mp.database_name,
                schema_name=mp.schema_name,
                table_name=mp.table_name,
                column_names=column_names
            ))
        
    def __extract_case_when_logic_row_filter__(self,sql_definition: str) -> RowFilterDetails:
            """
            Parse a CASE WHEN row filter definition to extract group memberships and their values.
            
            Args:
                sql_definition: The SQL CASE statement from a row filter
                
            Returns:
                Dictionary containing:
                - groups: List of dicts with 'group_name' and 'return_value'
                - default_value: The ELSE clause value
                - filter_type: The type of filter logic
            """
            result = RowFilterDetails(groups=[], default_value=None, row_filter_type=RowFilterType.EXPLICIT_GROUP_MEMBERSHIP)

            # Remove CASE and END
            content = sql_definition.replace('CASE', '').replace('END', '').strip()
            
            # Find all WHEN...THEN patterns
            when_pattern = r"WHENcurrent_role\(\)in\('([^']+)'\)THEN([^W]+?)(?=WHEN|ELSE|$)"
            when_matches = re.findall(when_pattern, content)
            
            # Find ELSE clause
            else_pattern = r"ELSE(.+?)$"
            else_match = re.search(else_pattern, content)
                        
            # Extract group and expression pairs
            for group, expression in when_matches:
                result.groups.append(RowFilterDetailGroup(group_name=group, return_value=expression.strip()))
            
            # Add ELSE clause if exists
            if else_match:
                result.default_value = else_match.group(1).strip()

            return result
            

    def __extract_if_logic_row_filter__(self, definition: str) -> RowFilterDetails:
        """
        Parse an IF row filter definition to extract group membership and their values.
        
        Args:
            sql_definition: The SQL IF statement from a row filter
        Returns:
            Dictionary containing:

        """
        
        group_name = definition[17:].split("'")[1]
        start_in = f"current_role()in('{group_name}'"
        equal = definition == f"current_role()=('{group_name}')"
        if not (definition.startswith(start_in) or equal):
            self.logger.warning("Unexpected format: does not match expected start")
            return result
        
        if equal:
            result = RowFilterDetails(
            groups=[RowFilterDetailGroup(
                group_name=group_name,
                return_value="true"
            )],
            default_value="false",
            row_filter_type=RowFilterType.EXPLICIT_GROUP_MEMBERSHIP)
            return result

        result = RowFilterDetails(
            groups=[],
            default_value="false",
            row_filter_type=RowFilterType.EXPLICIT_GROUP_MEMBERSHIP
        )
        group_names = definition[17:].split("'")
        for i, group_name in enumerate(group_names):
            if i % 2 == 0:
                continue
            else:
                rfdg = RowFilterDetailGroup(
                    group_name=group_name,
                    return_value="true"
                )
                result.groups.append(rfdg)

        return result


    def __extract_logic_from_row_filter__(self, sql_definition: str) -> RowFilterDetails:
        """
        Extracts the group name and mask pattern from a column mask function definition.
        
        Args:
            sql_definition (str): The SQL definition containing is_account_group_member function

        Returns:
            ColumnMaskExtraction: An object containing the extracted group name and mask pattern
        """

        result = RowFilterDetails(row_filter_type=RowFilterType.UNSUPPORTED)
        definition = sql_definition.replace("\n", " ").replace("\r", " ").replace(" ", "") 

        if definition.startswith("current_role()"):
            result_ = self.__extract_if_logic_row_filter__(definition)
        elif definition.startswith("CASEWHENcurrent_role()"):
            result_ = self.__extract_case_when_logic_row_filter__(definition)
        else:
            result_ = None
            
        
        if result_:
            return result_
        self.logger.warning("Unexpected format: does not start 'current_role()' or start with 'CASEWHENcurrent_role('")

        return result
    
    def __get_column_masks__(self, database: str) -> List[SnowflakeColumnMask]:
        """Retrieves the column masking policies for a specific database.
        Args:
            database (str): The name of the database for which to retrieve column masking policies.
        Returns:
            List[SnowflakeMaskingPolicy]: A list of SnowflakeMaskingPolicy objects representing the column masking policies.
        """

        query = f"""                            
                SELECT
                    mp.POLICY_BODY,
                    pr.POLICY_ID,
                    pr.POLICY_NAME,
                    pr.ref_database_name,
                    pr.ref_schema_name,
                    pr.ref_entity_name,
                    pr.ref_column_name
                FROM SNOWFLAKE.ACCOUNT_USAGE.POLICY_REFERENCES pr
                JOIN SNOWFLAKE.ACCOUNT_USAGE.MASKING_POLICIES mp
                    ON pr.POLICY_ID = mp.POLICY_ID
                WHERE pr.POLICY_KIND = 'MASKING_POLICY'
                AND pr.REF_DATABASE_NAME = '{database.upper()}' AND REF_ENTITY_DOMAIN = 'TABLE' AND POLICY_STATUS = 'ACTIVE';

                ;"""

        masking_policies = self.__run_query__(query, columns=["POLICY_BODY", "POLICY_ID",
                                                                   "POLICY_NAME", "ref_database_name",
                                                                   "ref_schema_name",
                                                                   "ref_entity_name",
                                                                   "ref_column_name"])
        
        
        self.masking_policies = list()
        self.tables_with_masks = list()
        for mp in masking_policies:
            extraction = self.__process_masking_policy__(mp["POLICY_BODY"])
            mp = SnowflakeMaskingPolicy(id=mp["POLICY_ID"],
                                        name=mp["POLICY_NAME"],
                                        database_name=mp["ref_database_name"],
                                        schema_name=mp["ref_schema_name"],
                                        table_name=mp["ref_entity_name"],
                                        column_name=mp["ref_column_name"],
                                        group_names=extraction.group_names,
                                        mask_pattern=extraction.mask_pattern,
                                        column_mask_type=extraction.column_mask_type
                                        )
            self.masking_policies.append(mp)
  
            query = f"""SELECT COLUMN_NAME FROM SNOWFLAKE.ACCOUNT_USAGE.COLUMNS WHERE TABLE_CATALOG = '{mp.database_name}' AND TABLE_SCHEMA = '{mp.schema_name}' AND TABLE_NAME = '{mp.table_name}';"""
            columns = self.__run_query__(query, columns=["COLUMN_NAME"])
            column_names = [c["COLUMN_NAME"] for c in columns]
            self.tables_with_masks.append(SnowflakeTableWithPolicy(
                database_name=mp.database_name,
                schema_name=mp.schema_name,
                table_name=mp.table_name,
                column_names=column_names
            ))
             

    def __process_masking_policy__(self, policy_body: str) -> SnowflakeColumnMaskExtraction:
        """
        Processes a masking policy dictionary to extract the mask type, group name, and mask pattern.
        Args:
            mp (dict): A dictionary representing a masking policy.
        Returns:
            SnowflakeColumnMask: A SnowflakeColumnMask object with the extracted information.
        """

        result = SnowflakeColumnMaskExtraction(column_mask_type=ColumnMaskType.UNSUPPORTED)

        definition = policy_body.replace("\n", " ").replace("\r", " ").replace(" ", "")

        if not(definition[:8] == "CASEWHEN" and definition[8:24] == "CURRENT_ROLE()IN"):
            self.logger.warning("Unexpected format: does not start with 'CASE WHEN CURRENT_ROLE() IN'")
            return result
        group_names = definition[25:].split(")")[0].replace("'", "").replace('"', '')

        index_ = 25 + len(group_names) + 3
        if definition[index_:index_+4] != "THEN":
            self.logger.warning("Unexpected format: 'THEN' not found where expected.")
            return result
        split_ = definition[index_ + 4 : ].split("ELSE")

        mask = None
        column_name_pos = None
        assigned_value = split_[0]
        if assigned_value[0] in ["'", '"']:
            assigned_value = assigned_value[1:-1]
            mask = assigned_value
        else:
            column_name_pos = 1

        unassigned_value = split_[1].replace("END", "")
        if unassigned_value[0] in ["'", '"']:
            unassigned_value = unassigned_value[1:-1]
            mask = unassigned_value
        else:
            column_name_pos = 2

        result.group_names = group_names.split(",")
        result.mask_pattern = mask
        if column_name_pos == 1:
            result.column_mask_type = ColumnMaskType.UNMASK_FOR_GROUP
        elif column_name_pos == 2:
            result.column_mask_type = ColumnMaskType.MASK_FOR_GROUP
        else:
            result.column_mask_type = ColumnMaskType.UNSUPPORTED

        
        return result



    def __get_user_role_assignment__(self, user: SnowflakeUserOrRole, is_user: bool) -> List[SnowflakeRole]:
        
        if is_user:
            directly_assigned = [role["NAME"] for role in self.user_assignments if role["GRANTEE_NAME"] == user.name]
        else:
            directly_assigned = [role["NAME"] for role in self.role_assignments if role["GRANTEE_NAME"] == user.name]

        assigned_roles = [role for role in self.roles if role.name in directly_assigned]

        # Inherited assignment
        for role in assigned_roles:
            inherited_roles = self.__get_user_role_assignment__(role, is_user=False)
            assigned_roles.extend(inherited_roles)

        return assigned_roles

    def __get_user_role_assignments__(self) -> dict[str, List[SnowflakeRole]]:
        user_role_assignments = dict()
        for user in self.users:
            user_role_assignments[user.name] = self.__get_user_role_assignment__(user, is_user=True)

        for role in self.roles:
            user_role_assignments[role.name] = self.__get_user_role_assignment__(role, is_user=False)

        return user_role_assignments

    def __get_role_membership__(self, role_name) -> Tuple[List[str], List[str]]:

        users = [user["GRANTEE_NAME"] for user in self.user_assignments if user["NAME"] == role_name]
        roles = list()

        for role_ass in self.role_assignments:
            if role_ass["NAME"] == role_name:
                roles.append(role_ass["GRANTEE_NAME"])

                users_, roles_ = self.__get_role_membership__(role_name=role_ass["GRANTEE_NAME"])
                users.extend(users_)
                roles.extend(roles_)

        return users, roles

    def __get_role_memberships__(self) -> dict[str, SnowflakeRoleMemberMap]:

        role_query = f"""select  "NAME", GRANTEE_NAME
                    from    SNOWFLAKE.ACCOUNT_USAGE.GRANTS_TO_ROLES
                    where   GRANTED_TO = 'ROLE' and
                            GRANTED_ON = 'ROLE' and
                            DELETED_ON is null and
                            PRIVILEGE = 'USAGE'"""

        self.role_assignments = self.__run_query__(role_query, columns=["NAME", "GRANTEE_NAME"])

        user_query = f"""select   "ROLE", GRANTEE_NAME 
                        from     SNOWFLAKE.ACCOUNT_USAGE.GRANTS_TO_USERS
                        WHERE DELETED_ON is null"""

        self.user_assignments = self.__run_query__(user_query, columns=["NAME", "GRANTEE_NAME"])

        role_memberships = dict()
        for role in self.roles:
            role_name = role.name
            users, roles = self.__get_role_membership__(role_name)

            member_users = [user for user in self.users if user.name in users]
            member_roles = [role for role in self.roles if role.name in roles]

            role_memberships[role_name] = SnowflakeRoleMemberMap(
                role_name=role_name,
                users=member_users,
                roles=member_roles
            )

        return role_memberships

    def __get_users__(self) -> List[SnowflakeUser]:

        columns = ["USER_ID", "NAME", "LOGIN_NAME", "EMAIL"]

        query = f"""SELECT {', '.join(columns)}
                    FROM SNOWFLAKE.ACCOUNT_USAGE.USERS where DELETED_ON is null and DISABLED = false
                    """

        users = self.__run_query__(query, columns=columns)
        return [SnowflakeUser(
            id=user["USER_ID"],
            name=user["NAME"],
            login_name=user["LOGIN_NAME"],
            email=user["EMAIL"]
        ) for user in users]

    def __get_roles__(self) -> List[SnowflakeRole]:
        """
        Retrieves all roles from the Snowflake account.
        Returns:
            List[SnowflakeRole]: A list of SnowflakeRole objects representing the roles in the account.
        """
        columns = ["ROLE_ID", "NAME"]

        query = f"""SELECT {', '.join(columns)}
                    FROM SNOWFLAKE.ACCOUNT_USAGE.ROLES where DELETED_ON is null and ROLE_TYPE = 'ROLE'

                    """

        roles = self.__run_query__(query, columns=columns)
        return [SnowflakeRole(
            id=role["ROLE_ID"],
            name=role["NAME"]
        ) for role in roles]
